/**********************************************************************
* File:			Vmpass.cpp
* DATE:     05-23-2006
* AUTHOR:   Lan Bai
* FUNCTION: find big global variables (>single_size), replace them with new
*           global variables defined by vm_malloc(size);
*						replace references to the big globals w/ function calls to
*           read_handle, write_handle;
*						insert call to initial() at the beginnig;
* To compile and use vmpass under LLVM 1.9:
*		under the directory of Vmpass.cpp: make ENABLE_OPTIMIZED=0
*		under the directory of code to be transformed (test.c):
*			opt -load=../../Debug/lib/LLVMVmpass.so -vmpass test.bc > test1.bc
*			llc -march=c test1.bc -of testm.c
*			llvm-gcc testm.c -o testm -L../../Debug/lib/memmu_rt.a
*
* CHANGELOG: 
*
* DATE:     06-15-2006
* AUTHOR:   Lan Bai
* CHANGE:   can correctly deal with computation based on one dimensional arrays
*
* DATE:     06-16-2006
* AUTHOR:   Lan Bai
* CHANGE:   can also deal with 2-D arrays accessed by [c][c], [i][c],[i][j].
*						c is constant. 
*						but how to deal with [c][i]? and more-than-2 D arrays? structures?
*						array passed as arguments?
*
* DATE:     06-20-2006
* AUTHOR:   Lan Bai
* CHANGE:   can also deal with [c][i] 
*						but still can't apply to convolution program, because the order
*						of deleting instructions is not right
*						def can't be erased before erasing its uses!
*						
* DATE:     06-21-2006
* AUTHOR:   Lan Bai
* CHANGE:   use reversed_iterator to iterase instmap in reversed order,
*						so all instructions are deleted in a correct order now
*						tested with convolution code, works correctly
*
* DATE:     05-02-2007
* AUTHOR:   Lan Bai
* CHANGE:   Some changes were made to work with LLVM-1.9. RegisterOpt -> 
*						RegisterPass; getValue() was ommited for ConstantInt; 
*						Pass.so is put into llvm_src/Release/lib/ after compilation. 
*						Don't link to LLVM when compile. 
*
* DATE:     05-14-2007
* AUTHOR:   Lan Bai
* CHANGE:   BaseTypeid was used incorrectly and caused seg fault. Fixed.
*
* DATE:     05-16-2007
* AUTHOR:   Lan Bai
* CHANGE:   Add read write function for signed types.
*
* DATE:     05-21-2007
* AUTHOR:   Lan Bai
* CHANGE:   Define cm_mem_area[] and pm_mem_area[] by vmpass. 
*
* DATE:     05-22-2007
* AUTHOR:   Lan Bai
* CHANGE:   Define all globals that are based on compressed and uncompressed
*						region sizes in Vmpass::ReplaceGlobals. 
*
* Fixme: 		How to maintain the initialized value of big globals?
*						why conflicting type for build-in func malloc?
***********************************************************************/

#include "Vmpass.h"

RegisterPass<Vmpass> X ("vmpass", "Virtual Mem Pass");

Vmpass::Vmpass() :
	total_page(10),
	single_size(20),
	page_size(256),
	comp_ratio(0.5),
	cm_page(0),
	pm_page(0),
	cm_size(0),
	pm_size(0),
	BigGlobals(),
	InstMap(),
	GepMap()
{}
 	
ModulePass * createVmPass() {
	return new Vmpass ();
}

bool Vmpass::runOnModule(Module & M) {
	bool Changed = false;
	const TargetData & TD = getAnalysis<TargetData>();
	Type::TypeID Basetypeid;

	if(SizeofGlobals(M,TD) <= total_page * page_size) {
		cerr << "memory usage is smaller than " << total_page 
			<< " pages, no need to use MEMMU" << endl;
		return Changed;
	}

	Changed = true;
	DivideMem();
  // loop over the module, add big globals with their sizes to BigGlobals
  for (Module::global_iterator GI = M.global_begin(), GE = M.global_end();
       GI != GE; ++GI) 
	{
		int gsize = TD.getTypeSize(GI->getType()->getElementType());
		if (gsize >= single_size) {
			ChangeGlobals(GI, gsize);	//add this global to BigGlobals
		}
	}

  //iterator all instructions, modify accesses to big globals
	for (Module::iterator FI = M.begin(), FE = M.end(); FI != FE; ++FI) {
		for (Function::iterator BI = FI->begin(), BE = FI->end(); BI != BE;
			++BI)
		{
			cerr << "---------------- Passing block " << BI->getName() << endl;
			for (BasicBlock::iterator II = BI->begin(), IE = BI->end();
	    	II != IE; ++II)
			{
				if (II->getOpcode() == Instruction::Load) {
					if (ConstantExpr * CE = dyn_cast<ConstantExpr>(II->getOperand(0))) {
						if (CE->getOpcode() == Instruction::GetElementPtr) {
							if (GlobalValue * GVa =
	      				dyn_cast<GlobalValue>(CE->getOperand (0)))
							{
								if (isBigGlobal(GVa)) {
									cerr << "************* load from big global " <<
										CE->getOperand(0)->getName () << endl;
									uint64_t voffset = getOffset(CE, &Basetypeid);
									cerr << "virtual offset is " << voffset << endl;
									Replaceload(&Basetypeid, II, GVa, voffset);
								}
		    			}
						}			
		    	}
				}
	      if (II->getOpcode () == Instruction::Store) {
//??? why constantexpr?
					if (ConstantExpr * CE =
		      	dyn_cast<ConstantExpr>(II->getOperand (1))) {
						if (CE->getOpcode() == Instruction::GetElementPtr) {
							if (GlobalValue * GVa =
								dyn_cast<GlobalValue>(CE->getOperand (0)))
							{
								if (isBigGlobal(GVa)){
									cerr << "************* store to big global "
			    					<< CE->getOperand (0)->getName () << endl;
									uint64_t voffset = getOffset(CE, &Basetypeid);
									cerr << "virtual offset is " << voffset <<
										endl;
									Replacestore(&Basetypeid, II, GVa, voffset);
								}
		    			}
						}
	    		}
				}
				if (II->getOpcode() == Instruction::GetElementPtr) {
					GetElementPtrInst *GEP = cast<GetElementPtrInst>(II);
					if(isa<GlobalVariable>(GEP->getPointerOperand())) {
						GlobalValue *GV =
							cast<GlobalValue>(GEP->getPointerOperand());
							if (isBigGlobal(GV)) {
								cerr << "********* " << GEP->getName() << 
									" index to bigglobal " <<	GV->getName () << endl;
								getGepOffset (GEP, &Basetypeid, TD);
							}
		    	}
		  		if (ConstantExpr * CE =
		      	dyn_cast<ConstantExpr>(GEP->getPointerOperand())) {
		      	if (CE->getOpcode () == Instruction::GetElementPtr){
							cerr << "gep pointer operand is constantexpr getelementptr"
			    			<< endl;
			  			if (GlobalValue * GVa =
			      		dyn_cast<GlobalValue>(CE->getOperand(0))) {
								if (isBigGlobal (GVa)) {
									cerr << "************ constantexpr gep access big global " 
							 			<< GVa->getName () << endl;
									GlobalValue *NG = findNewg(GVa);
									const Type *UnitType = cast<PointerType>(CE->getType())->getElementType();
									unsigned unitsize = TD.getTypeSize(UnitType);
 									uint64_t voffset = getOffset(CE, &Basetypeid);
      						Instruction * Gload = new LoadInst(NG, NG->getName() + ".ld", II);
     							Instruction *VAddrbase =
										BinaryOperator::create(Instruction::Add, Gload,
						 				ConstantInt::get(Type::UIntTy, voffset),
										GEP->getName () + ".base", GEP);
									Instruction * Cinst = new CastInst(GEP->getOperand(2), Type::UIntTy,
		    						GEP->getOperand(2)->getName() + ".uint", GEP);	
									Instruction * VMul = BinaryOperator::create(Instruction::Mul, Cinst,
										ConstantInt::get(Type::UIntTy,unitsize),
										GEP->getName () + ".mul", GEP);			
									Instruction * VAddr =
										BinaryOperator::create (Instruction::Add, VAddrbase,
						 				VMul, GEP->getName() + ".vaddr", GEP);
      						pair<Instruction *, Instruction *> Gpair;
      						Gpair.first = GEP;
      						Gpair.second = VAddr;
      						GepMap.insert (Gpair);
									cerr << "insert pair: " << GEP->getName() << ":" 
										<< VAddr->getName() << " to GepMap" << endl;
									ReplaceGepUses(II,TD);
								}
			    		}
						}
		    	}
				}
	   	}
		}
	}
  Replaceinsts (&M);
  Replaceglobals (&M);
  return Changed;
}

int Vmpass::SizeofGlobals(Module &M, const TargetData &TD) {
	int totalsize = 0;
	for (Module::global_iterator GI = M.getGlobalList().begin(), GE =
		M.getGlobalList().end(); GI != GE; ++GI){
		totalsize += TD.getTypeSize(cast<PointerType>(GI->getType())->getElementType());
	}
	GlobalSize = totalsize;
	required_size = totalsize;
	return totalsize;
}

// Divide the memory to compressed and uncompressed region. The sizes of them
// are computed based on compression ratio and total required size. 
void Vmpass::DivideMem() {
	required_page = (int)(required_size / page_size);
	cm_page = (int)(((total_page - required_page) / (1 - 1/comp_ratio)));
	pm_page = total_page - cm_page - 1;
	cm_size = page_size * cm_page;
	pm_size = page_size * pm_page;

}

bool Vmpass::isBigGlobal(GlobalValue * G) {
	map<GlobalValue *, NewGInfo>::iterator I = BigGlobals.find(G);
	if (I != BigGlobals.end()) {
		return 1;
  } else {
		return 0;
	}
}

//replace the gep inst with a series of insts to compute offset
//GEP is already confirmed to be accessing big globals
void Vmpass::getGepOffset(GetElementPtrInst * GEP, 
	Type::TypeID * Basetypeid, const TargetData &TD) {
	GlobalValue *OG = cast<GlobalValue>(GEP->getPointerOperand());
	GlobalValue *NG = findNewg(OG);
  //the final insruction of caluating offset,should be multiplication
	Instruction *Offset;
	Instruction *VAddr;
	const Type *BaseType = OG->getType ();
	//type of element this pointer reference
	const Type *UnitType = cast<PointerType>(GEP->getType())->getElementType();	
	unsigned unitsize = TD.getTypeSize(UnitType);
	const Type *Atype = cast<PointerType>(BaseType)->getElementType();
	cerr << "Atype: " << Atype->getDescription () << endl;
	for(unsigned i = 0; ArrayType::classof (Atype); i++) 
	{
		Atype = cast<ArrayType>(Atype)->getElementType();
	}
	*Basetypeid = Atype->getTypeID();
	unsigned base_size = Atype->getPrimitiveSize();
	cerr << "primitive size: " << base_size << endl;	
	//not sure yet, is this the only case when there is variable in index?
	if (GEP->getNumIndices() == 2) {
		Instruction *Cinst =
			new CastInst(GEP->getOperand(2), Type::UIntTy,
			GEP->getOperand (2)->getName() + ".uint", GEP);
			//this is the offset to the inner element
		Offset =
			BinaryOperator::create(Instruction::Mul, Cinst,
			ConstantInt::get(Type::UIntTy, unitsize),
			GEP->getName () + ".mult", GEP);
		Instruction *Gload = new LoadInst(NG, NG->getName () + ".ld", GEP);
		VAddr =
			BinaryOperator::create(Instruction::Add, Gload, Offset,
			GEP->getName () + ".addr", GEP);
		pair<Instruction *, Instruction *>Gpair;
		Gpair.first = GEP;
		Gpair.second = VAddr;
		GepMap.insert(Gpair);
		cerr << "insert pair: " << GEP->getName() << ":" 
			<< VAddr->getName() << " to GepMap" << endl;
	}else {
		cerr << "*^0^*,oops, did't expect index more than 2" << endl;
	}
	ReplaceGepUses(GEP,TD);
}

uint64_t Vmpass::getOffset(ConstantExpr * CE, Type::TypeID * Basetypeid) const{
	uint64_t offset = 0;
	uint64_t length[5] = {1, 1, 1, 1, 1};
	unsigned base_size = 0;
	cerr << "type " << CE->getOperand (0)->getType()->
		getDescription() << endl;
	if (PointerType::classof(CE->getOperand (0)->getType())) {
		const Type *Atype = cast< PointerType>
			(CE->getOperand(0)->getType())->getElementType();
		for (int i = 0; ArrayType::classof (Atype); i++)
		{
			length[i] = cast<ArrayType>(Atype)->getNumElements();
			Atype = cast<ArrayType>(Atype)->getElementType();
			cerr << "length[" << i << "]= " << length[i] << endl;
		}
		*Basetypeid = Atype->getTypeID();
		base_size = Atype->getPrimitiveSize();
		cerr << "size: " << base_size << endl;
	}

	for (unsigned i = 2; i < CE->getNumOperands (); i++) {
		if (cast<ConstantInt>(CE->getOperand (i))) {
		  cerr << "this index: " << cast<ConstantInt>(CE->getOperand(i)) << endl;
		  offset +=  cast<ConstantInt>(CE->getOperand(i))->getZExtValue() * length[i - 1];
			offset += 1;
		}
	}
	offset *= base_size;
	return offset;
}

void Vmpass::ChangeGlobals(GlobalValue * G, int bigglobalsize) {
	if(isBigGlobal (G)){
      return;
	}
	map<GlobalValue *, NewGInfo>::iterator I = BigGlobals.find(G);
  //create new global
	GlobalVariable *ngb = new
    GlobalVariable (Type::UIntTy, false, GlobalValue::ExternalLinkage,
    Constant::getNullValue (Type::UIntTy),
    G->getName () + "vm", G->getParent ());
	pair<GlobalValue *, NewGInfo>gp;
	gp.first = G;
	gp.second.NewGlobal = ngb;
	gp.second.GSize = bigglobalsize;
	BigGlobals.insert(I, gp);
	BigGlobalNum++;

	cerr << "********** big global " << G->getName () << " used in:" << endl;
	for (Value::use_iterator UI = G->use_begin(), UE = G->use_end(); UI != UE;
       ++UI)
	{
		(*UI)->dump();
		cerr << endl;
   }
}

GlobalValue *Vmpass::findNewg(GlobalValue * G) {
	map<GlobalValue *, NewGInfo>::iterator I = BigGlobals.find(G);
	if (I != BigGlobals.end ()) {
		return I->second.NewGlobal;
	}else {
		return 0;
	}
}

void Vmpass::Replacegmpload(Instruction * LI) {
	Module *M = LI->getParent()->getParent()->getParent();
	Instruction *RCall;
  //replace it with its virtual address
	map<Instruction *, Instruction *>::iterator NPair =
		GepMap.find (cast<Instruction>(LI->getOperand (0)));
	Instruction *NewI = NPair->second;
	// replace with call to read_handle
	vector<Value *>Args(1);
	Args[0] = NewI;
	Function *NRF;
	pair < Instruction *, Instruction * >INP;
	Type::TypeID Basetypeid = LI->getType()->getTypeID();
	switch(Basetypeid) {
		case Type::UByteTyID:
 			NRF = M->getOrInsertFunction 
				("read_handle8", Type::UByteTy, Type::UIntTy,	0);
			RCall = new CallInst(NRF, Args, "read_handle8", LI);
			cerr << "replaced load with call to read_handle8" << endl;
			break;
		case Type::SByteTyID:
 			NRF = M->getOrInsertFunction 
				("read_handle8s", Type::SByteTy, Type::UIntTy,	0);
			RCall = new CallInst(NRF, Args, "read_handle8s", LI);
			cerr << "replaced load with call to read_handle8s" << endl;
			break;
		case Type::UShortTyID:
			NRF =	M->getOrInsertFunction
				("read_handle16", Type::UShortTy, Type::UIntTy,	0);
			RCall = new CallInst(NRF, Args, "read_handle16", LI);
			cerr << "replaced load with call to read_handle16" << endl;
			break;
		case Type::ShortTyID:
			NRF =	M->getOrInsertFunction
				("read_handle16", Type::ShortTy, Type::UIntTy,	0);
			RCall = new CallInst(NRF, Args, "read_handle16s", LI);
			cerr << "replaced load with call to read_handle16s" << endl;
			break;
    case Type::UIntTyID:
			NRF =	M->getOrInsertFunction
				("read_handle32", Type::UIntTy, Type::UIntTy, 0);
			RCall = new CallInst(NRF, Args, "read_handle32", LI);
			cerr << "replaced load with call to read_handle32" << endl;
			break;
		case Type::IntTyID:
			NRF =	M->getOrInsertFunction
				("read_handle32s", Type::IntTy, Type::UIntTy, 0);
			RCall = new CallInst(NRF, Args, "read_handle32s", LI);
			cerr << "replaced load with call to read_handle32s" << endl;
			break;
		default:
			cerr << "default: " << Basetypeid << endl;
			break;
	}
	INP.first = LI;
	INP.second = RCall;
	InstMap.insert (INP);
	cerr << "inst map: " << INP.first->getName () << " ; " <<
		INP.second->getName () << endl;
	cerr << "type: " << INP.first->getType ()->getDescription () << " ; " <<
		INP.second->getType ()->getDescription () << endl;
	ReadNum++;
}

void Vmpass::Replacegmpstore(Instruction * SI) {
	Module *M = SI->getParent()->getParent()->getParent();
	Instruction *WCall;
	//replace it with its virtual address
	map < Instruction *, Instruction * >::iterator NPair =
		GepMap.find (cast < Instruction > (SI->getOperand (1)));
	Instruction *NewI = NPair->second;
	// replace with call to write_handle
	vector<Value *>Args (2);
	Args[0] = NewI;
	Args[1] = SI->getOperand(0);
	Function *NWF;
	Type::TypeID Basetypeid = SI->getOperand(0)->getType()->getTypeID();
	switch(Basetypeid) {
		case Type::UByteTyID:
			NWF =	M->getOrInsertFunction
				("write_handle8", Type::VoidTy, Type::UIntTy, Type::UByteTy, 0);
			WCall = new CallInst (NWF, Args, "", SI);
			SI->getParent ()->getInstList ().erase (SI);
			cerr << "replaced store with call to write_handle8" << endl;
			break;
		case Type::SByteTyID:
			NWF =	M->getOrInsertFunction
				("write_handle8s", Type::VoidTy, Type::UIntTy, Type::SByteTy, 0);
			WCall = new CallInst (NWF, Args, "", SI);
			SI->getParent ()->getInstList ().erase (SI);
			cerr << "replaced store with call to write_handle8s" << endl;
			break;
		case Type::UShortTyID:
			NWF =	M->getOrInsertFunction
				("write_handle16", Type::VoidTy, Type::UIntTy,
				Type::UShortTy, 0);
			WCall = new CallInst (NWF, Args, "", SI);
			SI->getParent ()->getInstList ().erase (SI);
			cerr << "replaced store with call to write_handle16" << endl;
			break;
		case Type::ShortTyID:
			NWF =	M->getOrInsertFunction
				("write_handle16s", Type::VoidTy, Type::UIntTy,
				Type::ShortTy, 0);
			WCall = new CallInst (NWF, Args, "", SI);
			SI->getParent ()->getInstList ().erase (SI);
			cerr << "replaced store with call to write_handle16s" << endl;
			break;
		case Type::UIntTyID:
			NWF =
				M->getOrInsertFunction ("write_handle32", Type::VoidTy, Type::UIntTy,
				Type::UIntTy, 0);
 			WCall = new CallInst(NWF, Args, "", SI);
			SI->getParent ()->getInstList ().erase (SI);
			cerr << "replaced store with call to write_handle32" << endl;
			break;
		case Type::IntTyID:
			NWF =
				M->getOrInsertFunction ("write_handle32s", Type::VoidTy, Type::UIntTy,
				Type::IntTy, 0);
 			WCall = new CallInst(NWF, Args, "", SI);
			SI->getParent ()->getInstList ().erase (SI);
			cerr << "replaced store with call to write_handle32s" << endl;
			break;
		default:
			cerr << "default: " << Basetypeid << endl;
			break;
	}
	WriteNum++;
}

void Vmpass::Replaceload(Type::TypeID * Basetypeid, Instruction * II,
		     GlobalValue * GVa, uint64_t voffset) {
	Module *M = II->getParent()->getParent()->getParent();
	Instruction *RCall;
	//replace it with its virtual address
	Instruction *Tempadd = 0;
	Value *OffsetVal = ConstantInt::get(Type::UIntTy, voffset);
	GlobalValue *Vglobal = findNewg(GVa);
	Instruction *Gload = 0;
	Gload = new LoadInst(Vglobal, GVa->getName () + ".ld", II);
	Tempadd =
		BinaryOperator::create (Instruction::Add, Gload, OffsetVal,
		II->getName () + ".add", II);
	// replace with call to read_handle
	vector < Value * >Args (1);
	Args[0] = Tempadd;
	Function *NRF;
	pair<Instruction *, Instruction *>INP;
  switch(*Basetypeid) {
		case Type::UByteTyID:
			NRF =
				M->getOrInsertFunction ("read_handle8", Type::UByteTy, Type::UIntTy,
				0);
			RCall = new CallInst (NRF, Args, "read_handle8", II);
			cerr << "replaced load with call to read_handle8" << endl;
			break;
		case Type::SByteTyID:
			NRF =
				M->getOrInsertFunction ("read_handle8s", Type::SByteTy, Type::UIntTy,
				0);
			RCall = new CallInst (NRF, Args, "read_handle8s", II);
			cerr << "replaced load with call to read_handle8s" << endl;
			break;
		case Type::UShortTyID:
			NRF =
				M->getOrInsertFunction ("read_handle16", Type::UShortTy, Type::UIntTy,
				0);
			RCall = new CallInst (NRF, Args, "read_handle16", II);
			cerr << "replaced load with call to read_handle16" << endl;
			break;
		case Type::ShortTyID:
			NRF =
				M->getOrInsertFunction ("read_handle16s", Type::ShortTy, Type::UIntTy,
				0);
			RCall = new CallInst (NRF, Args, "read_handle16s", II);
			cerr << "replaced load with call to read_handle16s" << endl;
			break;
		case Type::UIntTyID:
			NRF =
				M->getOrInsertFunction ("read_handle32", Type::UIntTy, Type::UIntTy,
				0);
			RCall = new CallInst (NRF, Args, "read_handle32", II);
			cerr << "replaced load with call to read_handle32" << endl;
 			break;
		case Type::IntTyID:
			NRF =
				M->getOrInsertFunction ("read_handle32s", Type::IntTy, Type::UIntTy,
				0);
			RCall = new CallInst (NRF, Args, "read_handle32s", II);
			cerr << "replaced load with call to read_handle32s" << endl;
 			break;
		default:
			cerr << "default: " << Basetypeid << endl;
			break;
	}
	INP.first = II;
	INP.second = RCall;
	InstMap.insert(INP);
	cerr << "inst map: " << INP.first->getName () << " ; " <<
		INP.second->getName() << endl;
	cerr << "type: " << INP.first->getType()->getDescription () << " ; " <<
		INP.second->getType()->getDescription() << endl;
	ReadNum++;
}

void Vmpass::Replacestore(Type::TypeID * Basetypeid, Instruction * II,
		      GlobalValue * GVa, uint64_t voffset) {
	Module *M = II->getParent()->getParent()->getParent();
	Instruction *WCall;
	//replace it with its virtual address
	Instruction *Tempadd = 0;
	Value *OffsetVal = ConstantInt::get(Type::UIntTy, voffset);
	GlobalValue *Vglobal = findNewg(GVa);
	Instruction *Gload = 0;
	Gload = new LoadInst(Vglobal, GVa->getName () + ".ld", II);
	Tempadd =
		BinaryOperator::create (Instruction::Add, Gload, OffsetVal,
		II->getName () + ".add", II);
	// replace with call to write_handle
	vector < Value * >Args (2);
	Args[0] = Tempadd;
	Args[1] = II->getOperand(0);
	Function *NWF;
	switch(*Basetypeid) {
		case Type::SByteTyID:
		case Type::UByteTyID:
			NWF =
				M->getOrInsertFunction("write_handle8", Type::VoidTy, Type::UIntTy,
				Type::UByteTy, 0);
			WCall = new CallInst(NWF, Args, "", II);
			II->getParent()->getInstList().erase(II);
			cerr << "replaced store with call to write_handle8" << endl;
			break;
		case Type::UShortTyID:
		case Type::ShortTyID:
			NWF =
				M->getOrInsertFunction("write_handle16", Type::VoidTy, Type::UIntTy,
				Type::UShortTy, 0);
			WCall = new CallInst(NWF, Args, "", II);
			II->getParent ()->getInstList ().erase (II);
			cerr << "replaced store with call to write_handle16" << endl;
			break;
		case Type::UIntTyID:
		case Type::IntTyID:
			NWF =
				M->getOrInsertFunction ("write_handle32", Type::VoidTy, Type::UIntTy,
				Type::UIntTy, 0);
			WCall = new CallInst (NWF, Args, "", II);
			II->getParent ()->getInstList ().erase (II);
			cerr << "replaced store with call to write_handle32" << endl;
			break;
		default:
			cerr << "default: " << Basetypeid << endl;
			break;
	}
	WriteNum++;
}

void Vmpass::Replaceinsts(Module * M) {
	for(map<Instruction *, Instruction * >::iterator MI =
		InstMap.begin (), ME = InstMap.end (); MI != ME; MI++) 
	{
		Instruction *GI = MI->first;
		for (Value::use_iterator UI = GI->use_begin(), UE = GI->use_end();
			UI != UE; UI++) {
		  cerr << "used in " << cast < Instruction >
	    (*UI)->getOpcodeName() << endl;
			for (unsigned i = 0; i < UI->getNumOperands(); i++) {
				if (UI->getOperand(i) == GI)	{
					cerr << "set operand " << i << " with " << 
						MI->second->getName() << endl;
					UI->setOperand(i, MI->second);
				}
			}
		}
	}
	//delete the original instructions
	for (Module::iterator FI = M->begin (), FE = M->end (); FI != FE; FI++) 
	{
		for (Function::iterator BB = FI->begin (), BE = FI->end (); BB != BE;
			BB++)
		{
			for (BasicBlock::iterator II = BB->begin (), IE = BB->end ();
				II != IE; II++) 
			{
				map < Instruction *, Instruction * >::iterator MI =
				InstMap.find (II);
				if (MI != InstMap.end ())	{
					cerr << "delete instruction: " << II->getName () << endl;
					BB->getInstList ().erase (II);
				}
			}
		}
	}
	//delete the original instructions of GepMap
	for(map<Instruction*,Instruction*>::reverse_iterator GEPI = GepMap.rbegin(), 
		GEPE = GepMap.rend(); GEPI != GEPE; ++GEPI) 
	{
			Instruction *ItoErase = GEPI->first;
			assert(ItoErase->getParent());
			cerr << "delete gep instruction: " << ItoErase->getName() << endl;
			ItoErase->getParent()->getInstList().erase(ItoErase);
	}
}

// Replace all big globals with new ones, insert call to vm_malloc for each
// insert call to initial at the beginning, and define globals that are used in
// MEMMU library.
void Vmpass::Replaceglobals(Module * M) {
  //delete the big globals
	map < GlobalValue *, NewGInfo >::iterator MI, EI;
	Function *FMain;
	for	(MI = BigGlobals.begin(), EI = BigGlobals.end(); MI != EI; MI++)
	{
		GlobalValue *Gv = MI->first;
		for (Module::global_iterator GI = M->getGlobalList().begin(), 
			GE = M->getGlobalList().end (); GI != GE; GI++)
		{
			if (GI == Gv) {
				cerr << "delete global:" << Gv->getName () << endl;
				M->getGlobalList().erase(GI);
			}
		}
		//insert calls to vm_malloc
		for (Module::iterator FI = M->begin(), FE = M->end(); FI != FE; FI++)
		{
			if (FI->getName () == "main"){
				FMain = FI;
			}
		}
		Function *MallocFn =
			M->getOrInsertFunction("vm_kma_malloc", Type::UIntTy, Type::UIntTy,
			0);
			std::vector<Value *>Args(1);
			Args[0] = ConstantInt::get(Type::UIntTy, MI->second.GSize);
			Instruction *IPoint = FMain->begin()->begin();
			Instruction *MCall = new CallInst(MallocFn, Args,
				MI->second.NewGlobal->getName() + ".malloc", IPoint);
			new StoreInst(MCall, MI->second.NewGlobal, IPoint);
			VmallocNum++;
	}

	//Define globals that are used in MEMMU library
	ConstantInt *const_cm = ConstantInt::get(Type::IntTy, cm_page);
	new GlobalVariable(Type::IntTy, true,
		GlobalValue::ExternalLinkage, const_cm, "MAX_CM_PAGE",
		M->getGlobalList().begin());

	ConstantInt *const_pm = ConstantInt::get(Type::IntTy, pm_page);
	new GlobalVariable(Type::IntTy, true,
		GlobalValue::ExternalLinkage, const_pm, "MAX_PM_PAGE",
		M->getGlobalList().begin());

	ConstantInt *const_vm = ConstantInt::get(Type::IntTy, required_page);
	new GlobalVariable(Type::IntTy, true,
		GlobalValue::ExternalLinkage, const_vm, "MAX_VM_PAGE",
		M->getGlobalList().begin());	

	ConstantInt *const_csize = ConstantInt::get(Type::IntTy, cm_size);
	new GlobalVariable(Type::IntTy, true,
		GlobalValue::ExternalLinkage, const_csize, "CM_SIZE",
		M->getGlobalList().begin());

	ConstantInt *const_psize = ConstantInt::get(Type::IntTy, pm_size);
	new GlobalVariable(Type::IntTy, true,
		GlobalValue::ExternalLinkage, const_psize, "PM_SIZE",
		M->getGlobalList().begin());	

	ConstantInt *const_vsize = ConstantInt::get(Type::IntTy, required_size);
	new GlobalVariable(Type::IntTy, true,
		GlobalValue::ExternalLinkage, const_vsize, "VM_SIZE",
		M->getGlobalList().begin());

	ArrayType *cm_ty = ArrayType::get(Type::UByteTy, cm_size);
	Constant *const_carea = ConstantAggregateZero::get(cm_ty);
	new GlobalVariable(cm_ty, false,
		GlobalValue::ExternalLinkage, const_carea, "cm_mem_area",
		M->getGlobalList().begin());

	ArrayType *pm_ty = ArrayType::get(Type::UByteTy, pm_size);
	Constant *const_parea = ConstantAggregateZero::get(pm_ty);
	new GlobalVariable(pm_ty, false,
		GlobalValue::ExternalLinkage, const_parea, "pm_mem_area",
		M->getGlobalList().begin());

	ArrayType *map_ty = ArrayType::get(Type::UByteTy, required_page);
	Constant *const_map = ConstantAggregateZero::get(map_ty);
	new GlobalVariable(map_ty, false,
		GlobalValue::ExternalLinkage, const_map, "map",
		M->getGlobalList().begin());

	ArrayType *fpage_ty = ArrayType::get(Type::UByteTy, pm_page + 1);
	Constant *const_fpage = ConstantAggregateZero::get(fpage_ty);
	new GlobalVariable(fpage_ty, false,
		GlobalValue::ExternalLinkage, const_fpage, "freepages",
		M->getGlobalList().begin());
	
  //insert initial
	Instruction *FI = FMain->begin()->begin();
	Function *InitFn = M->getOrInsertFunction ("initial", Type::VoidTy, 0);
	new CallInst (InitFn, "", FI);
}

void Vmpass::ReplaceGepUses(Instruction *II, const TargetData &TD) {
	GetElementPtrInst * GEP = cast<GetElementPtrInst>(II);
	for(Value::use_iterator UI = II->use_begin(), UE = II->use_end(); UI != UE;
		UI++)
	{
		cerr << "user of " << GEP->getName () << ": " 
			<< (*UI)->getName() << endl;
		if (Instruction * BI = dyn_cast<Instruction>(*UI)) {
			if (BI->getOpcode() == Instruction::Load) {
				cerr << "used in load: " << BI->getName() << endl;
				//replace with read_handle
				Replacegmpload (BI);
			}
			if (BI->getOpcode() == Instruction::Store) {
				cerr << "ReplaceGepUses: used in store" << endl;
				Replacegmpstore (BI);
			}
			if (BI->getOpcode() == Instruction::GetElementPtr) {
				cerr << "replacegepuses: used in getelementptr: " << BI->getName() << endl;
					Replacegmpgep (GEP,BI,TD);
			}
		}
	}
}

void Vmpass::Replacegmpgep(GetElementPtrInst *GEP,Instruction *BI,const TargetData &TD) 
{
	GetElementPtrInst *GEI = cast<GetElementPtrInst>(BI);
	if (GEI->getNumIndices() == 2) {   
		const Type *InnerUnitType = 
		cast < PointerType > (GEI->getType ())->getElementType();
		unsigned innersize = TD.getTypeSize(InnerUnitType);
		Instruction *CastII = new CastInst (GEI->getOperand (2), Type::UIntTy,
			GEI->getOperand (2)->getName () + ".uint", GEI);
		Instruction *LastOne = 
			BinaryOperator::create(Instruction::Mul, CastII, 
			ConstantInt::get(Type::UIntTy,innersize),GEI->getName()+".mul2",GEI);
		map<Instruction*,Instruction*>::iterator MI =	GepMap.find(GEP); 
		Instruction *FAdd = MI->second;
		Instruction *NewAddr = 
			BinaryOperator::create(Instruction::Add, FAdd, LastOne,GEI->getName()+".add2",GEI);
		pair < Instruction*, Instruction* > VPair;
		VPair.first = GEI;
		VPair.second = NewAddr;
		GepMap.insert(VPair);
		cerr << "insert pair: " << GEI->getName() << ":" 
				<< NewAddr->getName() << " to GepMap" << endl;
	}else {
		cerr << "*^0^*,can't deal with more than 3 indices" << endl;
	}
	for(Value::use_iterator USEI = GEI->use_begin(), USEE = GEI->use_end(); 
		USEI != USEE; USEI ++)
	{
 		if (Instruction * UBI = dyn_cast<Instruction>(*USEI)) {
 			if (UBI->getOpcode () == Instruction::Load) {
				cerr << "used in load: " << UBI->getName() << endl;
				Replacegmpload(UBI);
			}
 			if (UBI->getOpcode () == Instruction::Store) {
				cerr << "Replacegmpgep: used in store" << endl;
				Replacegmpstore (UBI);
			}
 			if (UBI->getOpcode () == Instruction::GetElementPtr) {	
				cerr << "used in getelementptr: " << UBI->getName() << endl;		
				Replacegmpgep(GEI,UBI,TD);			
			}
		}				
 	}	
}
