/*
 * Filename      : dcomp.c
 * Purpose       : Delta compression and decompression
 * Author        : Lei Yang
 * Date Created  : 03/28/2006
 * Last Modified : 04/09/2006. changed the statement to deal with
 *                             sign extending.
*/

/*****************************************************************************/
/*************************** Define and Includes *****************************/
//#define _DEBUG_

//#include <stdio.h>
#include <stdlib.h>
#include <string.h>
//#include <time.h>
//#include <math.h>
//#include <assert.h>
#include "dcomp.h"

#define TWO_BITS_PACKING_MASK 0x03030303
#define FOUR_BITS_PACKING_MASK 0x0f0f0f0f

#define BYTE_TWO_BITS_PACKING_MASK 0x03
#define BYTE_FOUR_BITS_PACKING_MASK 0x0f


#define fast_byte_copy(tgt, src) do { \
 *((unsigned char *)(tgt)) = *((unsigned char *)(src)); \
} while (0)

#define fast_short_copy(tgt, src) do { \
 *((uint16_t *)(tgt)) = *((uint16_t *)(src)); \
} while (0)

#define fast_word_copy(tgt, src) do { \
 *((int32_t *)(tgt)) = *((int32_t *)(src)); \
} while (0)


/*****************************************************************************/
/*************************** Function Declaration ****************************/
/* pack 2 bits bytes into bytes */
char *byte_pack_2bits(char *source_buf, char *source_end,
						char *dest_buf);

/* unpack 2 bits bytes from bytes */
char *byte_unpack_2bits(char *input_buf, char *input_end,
						char *output_buf);

/* pack 4 bits bytes into integral words */
char *byte_pack_4bits(char* source_buf, char* source_end,
	 	char* dest_buf);

/* unpack 4 bits bytes from integral words */
char *byte_unpack_4bits(char *input_buf, char *input_end,
	 	char *output_buf);

/*****************************************************************************/
/************************* Function Implementation ***************************/
/*****************************************************************************/
char *byte_pack_2bits(char *source_buf, char *source_end,
						char *dest_buf)
{
   register char *src_next = source_buf;
   char *dest_next = dest_buf;
  
   while (src_next < source_end) {
      register char temp = src_next[0];
      temp |= (src_next[1] << 2);
      temp |= (src_next[2] << 4);
      temp |= (src_next[3] << 6);
    
      dest_next[0] = temp;
      dest_next++;     
      src_next += 4;
   }
   return dest_next;
}

/*****************************************************************************/
char *byte_unpack_2bits(char *input_buf, char *input_end,
						char *output_buf)
{
  register char *input_next = input_buf;
  register char *output_next = output_buf;
  register char packing_mask = BYTE_TWO_BITS_PACKING_MASK;

  /* loop to repeatedly grab one input byte and unpack it into
   * 4 output bytes.  This loop could be unrolled a little---it's
   * designed to be easy to do that.
   */   
  while (input_next < input_end) {
    register char temp = input_next[0];
    output_next[0] = temp & packing_mask;
    output_next[1] = (temp >> 2) & packing_mask;
    output_next[2] = (temp >> 4) & packing_mask;
    output_next[3] = (temp >> 6) & packing_mask;
    
    output_next += 4;
    input_next++;
  }
  return output_next;
}

/*****************************************************************************/
char *byte_pack_4bits(char* source_buf, char* source_end,
	 	char* dest_buf)
{
	register char* src_next = source_buf;
	char* dest_next = dest_buf;
	
	/* this loop should probably be unrolled */
	while (src_next < source_end) {
		register char temp = src_next[0];
		temp |= (src_next[1] << 4);
		
		dest_next[0] = temp;
		dest_next++;     
		src_next += 2;
	}
	return dest_next;
}

/*****************************************************************************/
char *byte_unpack_4bits(char *input_buf, char *input_end,
	 	char *output_buf)
{
	register char *input_next = input_buf;
	register char *output_next = output_buf;
	register char packing_mask = BYTE_FOUR_BITS_PACKING_MASK;
	
	
	/* loop to repeatedly grab one input word and unpack it into
	* 4 output words.  This loop should probably be unrolled
	* a little---it's designed to be easy to do that.
	*/   
	while (input_next < input_end) {
		register char temp = input_next[0];
		//DEBUG_PRINT_2("Unpacked dictionary indices word: %.8x\n", temp);
		output_next[0] = temp & packing_mask;
		output_next[1] = (temp >> 4) & packing_mask;
		
		output_next += 2;
		input_next++;
	}
	return output_next;
}

/*****************************************************************************
 *                              Compressor                                   *
 *****************************************************************************/
void delta_compress(char *src, size_t srclen, 
										char *dest, size_t *destlen, 
										char *working_mem)
{
	/* working arrays */
	char tape1[TAPE_SIZE];
	char tape2[TAPE_SIZE];
	delta_word *tape3 = (delta_word *)(dest + sizeof(delta_word) + 
			PAGE_SIZE/(sizeof(delta_word) * 2) + PAGE_SIZE/(sizeof(delta_word) * 4));

	char *nexttape1 = tape1;
	char *nexttape2 = tape2;
	delta_word* nexttape3 = tape3;

	delta_word* thisword = (delta_word *)src;
	delta_word* lastword = NULL;

	char* next_dest = dest;

	int32_t diff = 0;

	int32_t num_uncompressed_words = srclen / sizeof(delta_word);

	/* If the page end up compressed, this value will change */
	*destlen = 0;

	fast_short_copy(next_dest, src);
	next_dest += sizeof(delta_word);

	while (num_uncompressed_words > 0)
	{
#ifdef _DEBUG_
		printf("The %d th delta_word to be compressed\n", 
			srclen / sizeof(delta_word) - num_uncompressed_words);
#endif
		/* very first delta_word */
		if (lastword == NULL)
		{
			diff = 0;
		}
		else
		{
			diff = (int32_t)(*thisword) - (int32_t)(*lastword);
		}
		
		if ( -31 <= diff && diff <= 31 )
		{
			/* first get the two bits */
			*nexttape2 = diff & 0x03;
			nexttape2++;

			/* then the higher four bits */
			*nexttape1 = (diff >> 2) & 0x0f;
			nexttape1++;
		}
		else
		{
			/* check if the compression is acceptable */
			if ( nexttape3 - tape3 == ACCEPTABLE_SIZE )
			{
				//fprintf(stderr, "compression result not acceptable!");
				return;
			}

			*nexttape1 = 0x08;
			nexttape1++;

			*nexttape2 = 0x00;
			nexttape2++;

			/* record the delta_word itself*/
			*nexttape3 = *thisword;
			nexttape3++;
		}	
		
		num_uncompressed_words--;
		lastword = thisword;
		thisword++;
	}

#ifdef _DEBUG_
	{int32_t i;
	for (i = 0; i < TAPE_SIZE; i++)
		printf("%x %x\n", tape1[i], tape2[i]);}
#endif
	/* pack the tapes */
	next_dest = (char *)byte_pack_4bits(tape1, nexttape1, next_dest);

	next_dest = (char *)byte_pack_2bits(tape2, nexttape2, next_dest);

	*destlen = (char *)nexttape3 - dest;

	return;
}


/*****************************************************************************
 *                              Deompressor                                  *
 *****************************************************************************/
void delta_decompress(char *src, size_t srclen,
											char *dest, size_t *destlen,
											char *working_mem)
{
	char tape1[TAPE_SIZE];
	char tape2[TAPE_SIZE];

	char *nexttape1 = src + sizeof(delta_word);
	char *nexttape2 = nexttape1 + PAGE_SIZE/(sizeof(delta_word) * 2);
	delta_word *nexttape3 = (delta_word *)(nexttape2 + PAGE_SIZE/(sizeof(delta_word) * 4));
	

	delta_word *next_dest = (delta_word *)dest;

	int32_t tapepos = 0;	
	int32_t diff = 0;

	delta_word lastword = *((delta_word *)src); /* get the very first word */
	delta_word thisword = 0;  // original
	
	/* unpack the tapes */
	byte_unpack_4bits(nexttape1, nexttape2, tape1);

	byte_unpack_2bits(nexttape2, (char *)nexttape3, tape2);

	nexttape1 = tape1;
	nexttape2 = tape2;

	for ( tapepos = 0; tapepos < TAPE_SIZE; tapepos++ )
	{
		/* get the bytes merge to a 6 bits code */
		diff = (*nexttape1) << 2;

		nexttape1++;
		diff = diff | (*nexttape2 & 0x03);

		nexttape2++;

		if ( (diff & 0x3f) == MAGIC_CODE )
		{
			thisword = *nexttape3;
			nexttape3++;
		}
		else
		{
			/* sign extend the 6bits */
			if (diff & 0x20) {
				diff |= 0xffffffc0;
			}

// **** this doesn't work for some processors ***
//			diff = diff << (32 - 6);
//			diff = diff >> (32 - 6);
// **********************************************
			thisword = (int32_t)lastword + diff;
		}
		*next_dest = thisword;
		next_dest++;
		lastword = thisword;
	}

	*destlen = (char *)next_dest - dest;

	return;
}

/********************************************************************/
#ifdef _DEBUG_

#define TEST_DATA_SIZE 8

int main()
{
	uint16_t testdata[TEST_DATA_SIZE]={
		0x02,
		0x03,
		0x04,
		0x05,
		0x06,
		0x07,
		0x08,
		0x09};

	unsigned char *ptr = NULL;

	char *dest = (char *)malloc(TEST_DATA_SIZE*sizeof(delta_word));
	int destlen;
	
	char *decompressed = (char *)malloc(TEST_DATA_SIZE*4);
	int decompressedlength;
	
	int i;
	int j = 0;
	
	delta_compress((char *)testdata, TEST_DATA_SIZE*sizeof(delta_word), dest, &destlen, 0);

	printf("comp_len %d\n", destlen);
	
	ptr = (unsigned char*)dest;
	for (i=0; i < destlen; i++)
		printf("%02x ", ptr[i]);

	printf("\n");

	delta_decompress(dest, destlen, decompressed, &decompressedlength, NULL);

	printf("decomp_len %d\n", decompressedlength);
	
	printf(">Before<     \t>After<\n");
	for (i = 0; i < TEST_DATA_SIZE; i++)
	{
		printf("%8x\t%8x",testdata[i],
			*((uint16_t *)(decompressed + i*2)));

		if (testdata[i] != *((uint16_t *)(decompressed+i*2)))
		{
			printf("  >>>>>>>>>>>>>>>>>>>>>>>");
		}
		printf("\n");
	}
	
	free(dest);
	free(decompressed);
	return 0;
}
#endif

/********************************************************************/

#ifdef _DEBUG_

int main(int argc, char *argv[])
{
	char *clean, *comp, *decomp;
	
	size_t size, comp_len, decomp_len;
	
	FILE *fp = NULL;
	int ret = 0;
	uint32_t count = 0, i = 0, notcompressed = 0;
	unsigned long total_comped = 0;
	unsigned long total = 0;
	clock_t c0, c1;
	float comp_time = 0, deco_time = 0;
	
	if (argc != 2)
		size = 256;
	else
		size = atoi(argv[1]);
	
	// Open the file first
	fp = fopen("../sound-example", "rb");
	if (fp == NULL)
	{
		perror("sound-example");
		return -1;
	}
	
	clean = (char *)malloc(size);
	comp = (char *)malloc(2*size);
	decomp = (char *)malloc(size);
	
	if (clean == NULL || decomp == NULL || comp == NULL)
	{
		printf("out of memory\n");
		return 3;
	}
	
	// Compress/decompress blocks until the whole file is done
	for (; count < 18*1024/size; count++)
	{
		ret = fread(clean, 1, size, fp);
		if (ret != size)
		{
			perror("sound-example");
			fclose(fp);
			return -1;
		}
		
		c0 = clock();
		for (i = 0; i < 10; ++i) {
			delta_compress(clean,size,comp,&comp_len, NULL);
		}
		c1 = clock();

		/* If this page is not compressed */
		if (comp_len == 0)
		{
			printf("page %d not compressed!\n", count);
			comp_time += (float)(c1-c0);
			total_comped += 10 * size;
			total += 10 * size;
			notcompressed ++;
			continue;
		}

		/* Else, the page is compressed, update the stats and decompress it */
		comp_time += (float)(c1-c0);
		total_comped += 10 * comp_len;
		total += 10 * size;
		
		c0 = clock(); 
		for (i = 0; i < 10; ++i) {
			delta_decompress(comp,comp_len,decomp,&decomp_len,NULL);
		}
		c1 = clock();
		deco_time += (float)(c1-c0);
		
		assert(size==decomp_len);

		ret = memcmp(clean,decomp,size);
		if (ret != 0)
		{
			printf(">Before<     \t>After<\n");
			for (i = 0; i < size; i+=4 )
			{
				printf("%8x\t%8x",*((uint16_t *)(clean+i)),*((uint16_t *)(decomp+i)));
				if (*((uint16_t *)(clean+i)) != *((uint16_t *)(decomp+i)))
				{
					printf("  >>>>>>>>>>>>>>>>>>>>>>>%d", count);
				}
				printf("\n");
			}
		}
		assert(ret==0);
	}

	comp_time /= CLOCKS_PER_SEC;
	deco_time /= CLOCKS_PER_SEC;
	
	fclose(fp);
	
	free(clean);
	free(comp);
	free(decomp);

	fprintf(stderr, "Block size (bytes)         : %4d\n", size);	
	fprintf(stderr, "Total compressed (bytes)   : %10ld\n", total_comped);
	fprintf(stderr, "Total read (bytes)         : %10ld\n", total);
	fprintf(stderr, "Average compression ratio  : %.8f\n", (float)total_comped/total);
	fprintf(stderr, "Average compression time   : %.8f\n", (float)comp_time/count/10);
	fprintf(stderr, "Average decompression time : %.8f\n", (float)deco_time/(count-notcompressed)/10);
	return 0;
}

#endif

