/*
 *  Filename:     kma_rm.c
 *  Purpose:      Inplement kma_rm
 *  Description:  Resource Map Allocator, we adopt "Best fit" policy here,
 *                always allocate memory from the smallest region that is large
 *                enough to satisfy the request. This has the drawback that it
 *                might leave several free regions that are too small to be useful.   
 * 
 *                Reserve the first physical page to store the informatio of memory objects:
 *                virtual address, size, next object in freelist.
 *
 *  CHANGELOG:
 *  
 *       Author:  Lei Yang
 *       Date:    2004-11-07
 *       Changes: Created
 *
 *       Author:  Lan Bai
 *       Date:    2005-12-25
 *       Changes: Change the return of kma_malloc and kma_free from a pointer to an offset from pm_start;
 *                we use one "page" in Lei's code as our whole virtual area.
 *                Ommit the argument "len" of kma_free, assume len always equals to the size of the object 
 *                to free;
 *
 *       Author:  Lan Bai
 *       Date:    2005-12-27
 *       Changes: Combine use_pool with use_freelist;
 *                Finish debugging, passed test programs.
 *                There is one thing need to improve: after free a memory object, if it is merged with 
 *                another freed object, what can we do with its objinfo? 
 *                First way: ignore this. But it causes wast of phy memory
 *                Seconde way: when the remaining reserved area is low,we begin to search for valid objifno
 *
 *       Author:  Lan Bai
 *       Date:    2006-01-30
 *       Changes: change the kma_malloc to deal with both decompressed area and compressed area
 *                using parameters to distinct these two situations
 *                how about using 1 and 0 to distinct, and make those parameters global?
 *
 *       Author:  Lan Bai
 *       Date:    2006-02-04
 *       Changes: considering compressed area and virtual area can have different number of memory objects,
 *                therefore introduce PM_INFO_RESERVE and CM_INFO_RESERVE(number of reserved pages) to 
 *                distinct them. 
 *                after a mem obj is freed, its res_fp's vaddr set to 0. so when res_fp went to the end
 *                search through res_fp, to find previous freed ones and reuse it.
 *       
 *       Author:  Lan Bai
 *       Date:    2006-02-05
 *       Changes: add pm_kma_malloc, and cm_kma_malloc. 
 *
 *       Author:  Lan Bai
 *       Date:    2006-03-01
 *       Changes: there were bugs when manage the freelist, some ajacent freed
 *                blocks were not efficiently merged, and the list was not in
 *                order. Which also caused seg fault. Now this problem is fixed.
 *
 *       Author:  Lan Bai
 *       Date:    2006-03-29
 *       Changes: divide the memory management of umcompressed and compressed
 *                area into different functions to improve the efficience of 
 *                the code. Add memory coalesce so fragmentation won't be a 
 *                problem. Further reduced the reserved area.
 *
 *       Author:  Lan Bai
 *       Date:    2006-03-31
 *       Changes: Fixed a bug in c_kma_free(). address should be assigned after
 *                fp.Another bug is in malloc, need to set cnext to NULL after
 *                malloc a block from compressed area. Besides,need to ajust 
 *                cnext for a objinfo when it is allocated.
 *     
 *       Author:  Lan Bai
 *       Date:    2007-05-20
 *       Changes: Let cm_mem_area[] and pm_mem_area[] be defined by vmpass.
 */
#include "kma_rm.h"

// uint8_t cm_mem_area[CM_SIZE];	// define by vmpass
uint8_t *cm_start = &cm_mem_area[0];  // pointer to compressed area

struct cm_objinfo *cm_flhead;
uint8_t cm_resfp = 0;
struct cm_objinfo cm_meminfo[MAX_CM_OBJ];

//uint8_t pm_mem_area[PM_SIZE];  // define by vmpass
uint8_t *pm_start = &pm_mem_area[0];  // pointer to uncompressed area

uint8_t cbuf_area[PAGE_SIZE];
uint8_t *cbuf = &cbuf_area[0];     
                               
struct vm_objinfo *vm_flhead;
uint8_t vm_resfp = 0;
struct vm_objinfo vm_meminfo[MAX_VM_OBJ];

#ifdef DEBUG
void print_cm_freelist();
#endif

#ifdef DEBUG
#define D(x) x
#else
#define D(X)
#endif

/************************* Function Implementation****************************/
#if 0
int main(void){ 

  vm_initial();
  cm_initial();
  
  uint16_t p1,p2,p3;
  uint8_t  q1,q2,q3,q4,q5;

  q1 = cm_kma_malloc(50);
  q2 = cm_kma_malloc(30);
  q3 = cm_kma_malloc(100);
  q4 = cm_kma_malloc(20);
  q5 = cm_kma_malloc(50);

  printf("q1 = %d, q2 = %d, q3 = %d, q4 = %d, q5 = %d\n",q1,q2,q3,q4,q5);
  c_kma_free(q2);
  c_kma_free(q4);
  
  q2 = cm_kma_malloc(40);
  c_kma_free(q1);
  q1 = cm_kma_malloc(10);
  q4 = cm_kma_malloc(52);

}
#endif


/*************************** Function Declaration ****************************/
struct cm_objinfo * c_rm_exact_fit(uint16_t len, uint16_t *sz, struct cm_objinfo **fp,
			 												struct cm_objinfo **fpp);
struct cm_objinfo * c_rm_best_fit(uint16_t len, uint16_t sz, struct cm_objinfo *fp,
			 												struct cm_objinfo *fpp);
struct cm_objinfo * c_kma_malloc(uint16_t size);

struct vm_objinfo * v_rm_exact_fit(uint16_t len, uint16_t *sz, struct vm_objinfo **fp,
			 												struct vm_objinfo **fpp);
struct vm_objinfo * v_rm_best_fit(uint16_t len, uint16_t sz, struct vm_objinfo *fp,
			 												struct vm_objinfo *fpp);
struct vm_objinfo * v_kma_malloc(uint16_t size);

void cm_coalesce();

/****************************************************************************/
void cm_initial() {
	uint8_t i;
  for(i = 0; i < MAX_CM_OBJ; ++i) {
		cm_meminfo[i].caddr = 0;
		cm_meminfo[i].csize = 0;
    cm_meminfo[i].cnext = 0;
	}
 	cm_flhead = &cm_meminfo[0];
	cm_flhead->csize = CM_SIZE;
	cm_flhead->caddr = cm_start;
  cm_flhead->cnext = cm_flhead;
  cm_resfp = 1;
} 

/*****************************************************************************/
struct cm_objinfo * c_kma_malloc (uint16_t len) {

  struct cm_objinfo *fpbestfit = 0;   /* best_fit free chunk */
  struct cm_objinfo *fpbestpre = 0;   /* the one before best_fit chunk */
  struct cm_objinfo *ret = 0;         /* return value */
  uint16_t sz = 0;                /* size of smallest fit chunk */
  
  D(printf("                                                               "));
  D(printf("cm_malloc %4d, ",len));

  /* 
   * First, walk through the free list and try to find a chunk that could match
   * exactly. If we succeed, we are done and return. While searching along the
   * list, we keep record of size of smallest chunk we've found that would
   * still fit the request -- we need it for step 2.
   */
  ret = c_rm_exact_fit (len, &sz, &fpbestfit, &fpbestpre);
  if (ret != 0) {
    return ret;
  }
      
  /*
   * Step2: If we found a chunk in the objinfo that could fit, but was larger
   * than asked for. Look it up again and use it. Since the objinfo entry now
   * need to be split into two now, watch out that the difference between the
   * requested size and the size of the chunk found need to be large enough
   * for another objinfo entry; if not, just enlarge the request size to what
   * we've found, and use the entire chunk.
   */
      
  if (sz) {
    ret = c_rm_best_fit (len, sz, fpbestfit, fpbestpre);
    return ret;
  }

  /*
   * Step3: We've searched through all pages we have, no one could provide us
   * a big enough chunk. 
   */
  D(printf ("failed\n"));
    
  return 0;
}
/*****************************************************************************/
struct cm_objinfo * c_rm_exact_fit (uint16_t len, uint16_t *sz, 
			       												struct cm_objinfo **fp, 
																		struct cm_objinfo **fpp) {

  struct cm_objinfo *fp1, *fp2;
  
  for(fp1 = cm_flhead, fp2 = 0; (fp1 != cm_flhead) || (fp2 == 0); fp2 = fp1, fp1 = fp1->cnext) {
    /* Found it! Remove this chunk from objinfo and return it.*/
    if (fp1->csize == len) {
      if (fp2) {
        fp2->cnext = fp1->cnext;
				fp1->cnext = 0;
      }
      else {
				for(fp2 = cm_flhead; fp2; fp2 = fp2->cnext) {
					if(fp2->cnext == cm_flhead) {
						fp2->cnext = fp1->cnext;
						break;
					}
				}
        cm_flhead = fp1->cnext;
				fp1->cnext = 0;
      }
      D (printf ("virtual addr: %d\n",fp1->caddr - cm_start));
      return fp1;
    }
    /*
     * If the chunk size scanned this time is larger than required,
     * we keep record of the smallest chunk met so far.
     */
    if (fp1->csize > len) {
      if ( *sz == 0 || fp1->csize < *sz) {
	      *sz = fp1->csize;
				*fp = fp1;
				*fpp = fp2;
      }
    }
  }
  /* We reach here if we couldn't find a exact_fit chunk */
  return 0;
}
/*****************************************************************************/
struct cm_objinfo * c_rm_best_fit (uint16_t len, uint16_t sz, struct cm_objinfo *fpbest,
			      										struct cm_objinfo *fpbestpre) {
  struct cm_objinfo *fp1;
  uint16_t length = len;
  uint16_t size = sz;
  struct cm_objinfo *fp = fpbest;
  struct cm_objinfo *fpp = fpbestpre;
  uint8_t avail_resfp;	 	

  /* Use the entire chunk */
  if(size - length <= 4) {
    if (fpp) {
      fpp->cnext = fp->cnext;
			fp->cnext = 0;
    } else {
			for(fp1 = cm_flhead; fp1; fp1 = fp1->cnext) {
				if(fp1->cnext == cm_flhead) {
					fp1->cnext = fp->cnext;
					break;
				}
			}					
      cm_flhead = fp->cnext;
			fp->cnext = 0;
    }
    D (printf ("virtual addr: %d\n",fp->caddr - cm_start));
    return fp;
  }
  
  /*
   * Else, split this chunk into two. Note that we leave the first
   * part as the new (smaller) entry, and return the uppper
   * portion to the caller. This saves us the work to fix up the
   * objinfo chain; we just need to fix up the size of the current
   * entry, and note down the size of the new chunk before
   * returning it to the caller.
   */
  //first, check whether we have space for a new objinfo
  if(cm_resfp >= MAX_CM_OBJ) {
    //find an availabe objinfo unit from reserved area
    for(avail_resfp = 0; avail_resfp < MAX_CM_OBJ; ++avail_resfp) {
      if(cm_meminfo[avail_resfp].caddr == 0) {
				break;
      }
    }
    if(avail_resfp == MAX_CM_OBJ + 1) {
      D(printf("cm_malloc failed, no space for objinfo!\n"));
      return 0;
    } else {
      fp1 = &cm_meminfo[avail_resfp];
    }
  } else {
    fp1 = &cm_meminfo[cm_resfp];
    ++ cm_resfp;
  }
  fp1->caddr = fp->caddr;
  fp1->csize = length;
  fp1->cnext = 0;
  
  size -= length;
  fp->csize = size;
  fp->caddr += length;

  D (printf ("virtual addr: %d\n",fp1->caddr - cm_start));
  return fp1;
}
/*****************************************************************************/
void c_kma_free (uint8_t freeid) {
  struct cm_objinfo *fp,*fp1;
  uint8_t *address = 0;
  fp = &cm_meminfo[freeid];
  address = fp->caddr;

  D(printf("                                                               "));
  D(printf("cm_free: %d\n",fp->caddr - cm_start));

  //locate the insert point in freelist
  if(fp->caddr < cm_flhead->caddr) {  
  // insert the newly freed object into the head of the freelist
     if(fp->caddr + fp->csize == cm_flhead->caddr) {
        cm_flhead->caddr = fp->caddr;
        cm_flhead->csize += fp->csize;
        fp->caddr = 0;
        fp->csize = 0;
        fp->cnext = 0;
     } else {
     		fp->cnext = cm_flhead;
				for(fp1 = cm_flhead; ; fp1 = fp1->cnext) {
					if(fp1->cnext == cm_flhead) {
						fp1->cnext = fp;
						break;
					}
				}
     		cm_flhead = fp;
     }
  } else {
  for(fp1 = cm_flhead; fp1 != 0; fp1 = fp1->cnext) {
    if(fp1->cnext != cm_flhead) {
      if(fp1->caddr < address && fp1->cnext->caddr > address) {
			//merge the free object with the prevois one
			  if(fp1->caddr + fp1->csize == fp->caddr) {
	  		  fp1->csize += fp->csize;
	  		  fp->caddr = 0;
	  		  fp->csize = 0;
          if(fp1->caddr + fp1->csize == fp1->cnext->caddr) {
             fp1->csize += fp1->cnext->csize;
             fp1->cnext->caddr = 0;
             fp1->cnext->csize = 0;
             if(fp1->cnext->cnext) {
               fp1->cnext = fp1->cnext->cnext;
             } else {
               fp1->cnext = 0;
             }
           }
	  		  break;
			  }
      //merge the free object with the following one
		if(fp->caddr + fp->csize == fp1->cnext->caddr) {
	  	fp1->cnext->caddr = fp->caddr;
	  	fp1->cnext->csize += fp->csize;
	  	fp->caddr = 0;
	  	fp->csize = 0;
	  	break;
		}
		//insert the free object into the freelist
	   fp->cnext = fp1->cnext;
	   fp1->cnext = fp;
     break;
     }
    } else {     
       //append it to the tail of the list
      fp1->cnext = fp;
      fp->cnext = cm_flhead;
      break;
    }
  }
 }
#ifdef DEBUG
  print_cm_freelist();
#endif
  return;
}
/******************************************************************************/
uint8_t cm_kma_malloc(ad_size_t len) {
  struct cm_objinfo *cm_fp = c_kma_malloc(len); 
  if(cm_fp) {
#ifdef DEBUG
    print_cm_freelist();
#endif
    return (uint8_t)(cm_fp - &cm_meminfo[0]);
  } else {
			cm_coalesce();
    	cm_fp = c_kma_malloc(len);
    	if(cm_fp) {
				D(printf("-------------------implement coalesce----------------------\n"));
#ifdef DEBUG
				print_cm_freelist();
#endif
      	return (uint8_t)(cm_fp - &cm_meminfo[0]);
			} else {
    		D(printf("cm_kma_malloc failed\n"));
    		exit(0);
    	}
  	}
}
/*****************************************************************************/
/*****************************************************************************/
void vm_initial() {
	uint8_t i;
  for(i = 0; i < MAX_VM_OBJ; ++i) {
		vm_meminfo[i].vaddr = 0;
		vm_meminfo[i].vsize = 0;
    vm_meminfo[i].vnext = 0;
	}
 	vm_flhead = &vm_meminfo[0];
	vm_flhead->vsize = VM_SIZE;
	vm_flhead->vaddr = 0;
  vm_resfp = 1;
	printf("initial done\n");
} 

/********************************************************************************/
struct vm_objinfo * v_rm_exact_fit (const uint16_t len, uint16_t *sz, 
			       										struct vm_objinfo **fp, struct vm_objinfo **fpp) {

  struct vm_objinfo *fp1, *fp2;
  
  for (fp1 = vm_flhead, fp2 = 0; fp1; fp2 = fp1, fp1 = fp1->vnext) {
    /* Found it! Remove this chunk from objinfo and return it.*/
    if (fp1->vsize == len) {
      if (fp2) {
        fp2->vnext = fp1->vnext;
      }
      else {
        vm_flhead = fp1->vnext;
      }
      D (printf ("virtual addr: %d\n",fp1->vaddr));
      return fp1;
    }
    /*
     * If the chunk size scanned this time is larger than required,
     * we keep record of the smallest chunk met so far.
     */
    if (fp1->vsize > len) {
      if ( *sz == 0 || fp1->vsize < *sz) {
	      *sz = fp1->vsize;
				*fp = fp1;
				*fpp = fp2;
      }
    }
  }
  /* We reach here if we couldn't find a exact_fit chunk */
  return 0;
}

/*****************************************************************************/
struct vm_objinfo * v_rm_best_fit (uint16_t len, uint16_t sz, struct vm_objinfo *fpbest,
			      									struct vm_objinfo *fpbestpre) {
  struct vm_objinfo *fp1;
  uint16_t length = len;
  uint16_t size = sz;
  struct vm_objinfo *fp = fpbest;
  struct vm_objinfo *fpp = fpbestpre;	 	
  uint8_t avail_res_fp;
  /* Use the entire chunk */
  if(size - length < 4) {
    if (fpp) {
      fpp->vnext = fp->vnext;
    } else {
      vm_flhead = fp->vnext;
    }
    D (printf ("virtual addr: %d\n",fp->vaddr));
    return fp;
  }
  
  /*
   * Else, split this chunk into two. Note that we leave the first
   * part as the new (smaller) entry, and return the uppper
   * portion to the caller. This saves us the work to fix up the
   * objinfo chain; we just need to fix up the size of the current
   * entry, and note down the size of the new chunk before
   * returning it to the caller.
   */
  //first, check whether we have space for a new objinfo
  if(vm_resfp >= MAX_VM_OBJ) {
    //find an availabe objinfo unit from reserved area
    for(avail_res_fp = 0;avail_res_fp < MAX_VM_OBJ; ++avail_res_fp) {
      if(vm_meminfo[avail_res_fp].vaddr == 0) {
//				D(printf("                    find valid objinfo unit from previous freed ones\n"));
				break;
      }
    }
    if(avail_res_fp == MAX_VM_OBJ + 1) {
//      printf("vm_malloc failed, no space for objinfo!\n");
      return 0;
    } else {
      fp1 = &vm_meminfo[avail_res_fp];
    }
  } else {
    fp1 = &vm_meminfo[vm_resfp];
    ++ vm_resfp;
  }
  fp1->vaddr = fp->vaddr;
  fp1->vsize = length;
  fp1->vnext = 0;
  
  size -= length;
  fp->vsize = size;
  fp->vaddr += length;

  D (printf ("virtual addr: %d\n",fp1->vaddr));
  return fp1;
}

/*********************************************************************************/
struct vm_objinfo * v_kma_malloc (uint16_t len) {

  struct vm_objinfo *fpbestfit = 0;   /* best_fit free chunk */
  struct vm_objinfo *fpbestpre = 0;   /* the one before best_fit chunk */
  struct vm_objinfo *ret = 0;             /* return value */
  uint16_t sz = 0;                /* size of smallest fit chunk */
  
  D(printf("                                                               "));
  D(printf("vm_malloc %4d, ",len));

  /* We cannot allocate an object larger than avail */
  if (len > VM_SIZE) {
    D(printf("can't malloc larger than %d\n", VM_SIZE));
    return 0;
  }

  /* 
   * First, walk through the free list and try to find a chunk that could match
   * exactly. If we succeed, we are done and return. While searching along the
   * list, we keep record of size of smallest chunk we've found that would
   * still fit the request -- we need it for step 2.
   */
  ret = v_rm_exact_fit (len, &sz, &fpbestfit, &fpbestpre);
  if (ret != 0) {
    return ret;
  }
      
  /*
   * Step2: If we found a chunk in the objinfo that could fit, but was larger
   * than asked for. Look it up again and use it. Since the objinfo entry now
   * need to be split into two now, watch out that the difference between the
   * requested size and the size of the chunk found need to be large enough
   * for another objinfo entry; if not, just enlarge the request size to what
   * we've found, and use the entire chunk.
   */
      
  if (sz) {
    ret = v_rm_best_fit (len, sz, fpbestfit, fpbestpre);
    return ret;
  }

  /*
   * Step3: We've searched through all pages we have, no one could provide us
   * a big enough chunk. 
   */
  D(printf ("failed\n"));
    
  return 0;
}

/*****************************************************************************/
uint16_t vm_kma_malloc(ad_size_t len) {
  struct vm_objinfo *vm_fp = v_kma_malloc(len);
  if(vm_fp) {
    return vm_fp->vaddr;
  } else {
    D(printf("vm_kma_malloc failed\n"));
    exit(0);
  }
}  

/*****************************************************************************/
void cm_coalesce() {
	uint8_t *start_p = cm_start;
  uint8_t *minaddr = 0;
  uint8_t i,j,min;
  struct cm_objinfo *fp1,*fp2;
  struct cm_objinfo *minfp = &cm_meminfo[0];

  min = 0;
	for(j = 0; j < MAX_CM_OBJ; ++j) {
   minaddr = 0;
 	 for(i = 0; i < MAX_CM_OBJ; ++i) {
			if((cm_meminfo[i].caddr >= start_p) &&(cm_meminfo[i].cnext == 0)) {
				if(minaddr == 0) {
					minaddr = cm_meminfo[i].caddr;
				}
    		if(cm_meminfo[i].caddr <= minaddr) {
				 minaddr = cm_meminfo[i].caddr;
         min = i;
    		}
  	}
	}
	if(minaddr == 0) {
		break;
	} else {
  	minfp = &cm_meminfo[min];
  	if(minfp->caddr > start_p) {
			memcpy(start_p,minfp->caddr,minfp->csize);
			minfp->caddr = start_p;
    	start_p += minfp->csize;
  	} else {
			start_p += minfp->csize;
		}
	}
 }
//update freelist
 for(fp1 = cm_flhead->cnext,fp2 = cm_flhead; fp1; fp2 = fp1,fp1 = fp1->cnext) {
		fp2->caddr = 0;
		fp2->csize = 0;
		fp2->cnext = 0;
		if(fp1 == cm_flhead) {
			fp1->caddr = start_p;
			fp1->csize = cm_start + CM_SIZE - start_p;
			fp1->cnext = cm_flhead;
      break;
		}
	}
}

/*********************************************************************************************/

#ifdef DEBUG
void print_cm_freelist() {
   struct cm_objinfo *fobj;
   printf("                                                cm freelist is : \n");
   printf("                                                           ");
   for(fobj = cm_flhead; fobj->cnext != cm_flhead; fobj = fobj->cnext) {
       printf(" - > (%d, %d)",fobj->caddr - cm_start,fobj->csize);
   }
   printf(" - > (%d, %d)",fobj->caddr - cm_start, fobj->csize);
   printf("\n");
}
#endif

//#endif /*  KMA_RM*/
