#!/usr/bin/env python

#----------------------------------------------------------------------------#
# Script name: apply_memmu.py
# Author:      Lan Bai (l-bai@northwestern.edu)
#
# This script automaicly apply the MEMMU transformation and optimization flow
# on the input C program. Input the name of the C program, for example, use
# "test" as the argument if the orginal program is "test.c". Then this script 
# will generate the following files:
#
# test.c --> (asm_pre) --> test_noasm.c --> (llvm-gcc) --> test_noasm.bc -->
#	(MEMMU transformation) --> test_noasm_memmu.bc --> (asm_post) -->
# test_noasm_memmu.c --> (LLVM C-backend) --> test_memmu.c
# 
# You need to have LLVM, MEMMU library and MEMMU passes installed
# before running this script.
#----------------------------------------------------------------------------#

import sys, re, os

class Dmap(object):
	'''Dmap is a special map structure that both the key and value are unique and
	can be searched by key (ffind) or value (bfind).'''
	def __init__(self):
		self.fmap = {}
		self.bmap = {}
	def ffind(self, k):
		return self.fmap[k]
	def bfind(self, k):
		return self.bmap[k]
	def add(self, (k,v)):
		if not k in self.fmap:
			self.fmap[k] = v
			self.bmap[v] = k
	def has(self, k):
		return k in self.fmap

def asm_pre(ifname, ofname):
	'''Replace "__asm volatile(...);" in file.c with asm_replace_func(id) and generate
	ofile, meanwhile build a mapping from id to the origial asm instructions.
	Hack: assume there can not be multiple __asm volatile in the same line. '''
	ifile = file(ifname, 'rb')
	data = ifile.readlines()
	ifile.close()
	count = 0
	ofile = file(ofname, 'wb')
	for i in range(len(data)):
		ln = data[i]
		m = re.match('\s+__asm volatile.+$', ln)
		if m:
			if ln.find(');') >= 0:
#				print 'ln: ', ln
				pass
			else:
				while 1:
					ln = ln + data[i+1]
					i += 1
					if ln.find(');') >= 0:
#						print 'ln: ', ln
						break
			mm = re.match('\s+__asm volatile\s+\((.+)\);(.*)$', ln)
			if mm:
				asm = mm.groups()[0]
				other = mm.groups()[1]
				if asm_map.has(asm):
					id = asm_map.ffind(asm)
				else:
					count += 1
					asm_map.add((asm, count))
					id = count
				ofile.write('asm_replace_func(' + str(id) + ');\n')
				ofile.write(other + '\n')
		else:
			ofile.write(ln)
	ofile.close()
	return

def asm_post(ifname, ofname):
	'''Replace asm_replace_func() with the original inline asm.''' 
	ifile = file(ifname, 'rb')
	data = ifile.readlines()
	ifile.close()
	ofile = file(ofname, 'wb')

	for ln in data:
		m = re.match('.+asm_replace_func\((.+)\);$', ln)
		if m:
			id = m.groups()[0]
			asm = asm_map.bfind(int(id))
			asm = '\t' + '__asm volatile (' + asm + ');\n'
#			print 'asm: ', asm
			ofile.write(asm)
		else:
			ofile.write(ln)
	ofile.close()
	return

if __name__ == '__main__':
	if len(sys.argv) != 2:
		sys.exit('usage: asm_process.py [file name]')

	fname = sys.argv[1]
	fname1 = fname + '.c'
	fname2 = fname + '_noasm.c'
	fname3 = fname + '_noasm.bc'
	fname4 = fname + '_noasm_memmu.bc'
	fname5 = fname + '_noasm_memmu.c'
	fname6 = fname + '_memmu.c'

	asm_map = Dmap()

	asm_pre(fname1, fname2)

	command = 'llvm-gcc -Wa,disable_opt -emit-llvm -c ' + fname2 + \
		' -o ' + fname3
	print command
	os.system(command)

	command = 'opt -load=../../Debug/lib/LLVMVmpass.so -vmpass ' + \
		fname3 + ' > '+ fname4
	print command
	os.system(command)

	command = 'llc -march=c ' + fname4 + ' -f -o ' + fname5
	print command
	os.system(command)

	asm_post(fname5, fname6)

	command = 'rm ' + fname2 + ' ' + fname3 + ' ' + fname4 + ' ' + fname5
	print command
	os.system(command)

	print '******** MEMMU flow is done. Output file is: ', fname6

